%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Script to find the optimal assembly sequence and task allocation for a workstation
% with one collaborative robot and one human worker considering completion
% time and ergonomic stress level using multi-agent reinforcement learning
%
% The model is part of the Master thesis "Assembly Line Sequencing with Collaborative 
% Robots using Reinforcement Learning" by Luisa Alina Kinat from June 2023
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear
clc
close all

% When doTraining = false, already trained agents can be used for simulation
doTraining = false;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Define variables
num_tasks = 7; % Total number of tasks
max_time = 60*60; % Maximum time per task
max_erg = 100; % Maximum ergonomic stress level per task

% Create a vector containing all possible actions
possible_actions = zeros(1,num_tasks+1);
for i = 0:num_tasks
    possible_actions(1,i+1) = i;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Define the action space (later used to create the RL agents)
% Number of task which the agent chooses to do
act(1) = Simulink.BusElement;
act(1).Name = "action";
act(1).Dimensions = 1;

% Create bus object
actBus = Simulink.Bus();
actBus.Elements = act;
actInfo = bus2RLSpec("actBus","DiscreteElements",{"action",possible_actions});

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Define the observation space (later used to create the RL agents)
% Number of task which the robot-agent is currently doing (discrete)
obs(1) = Simulink.BusElement;
obs(1).Name = "isDoingA";

% Number of task which the human-agent is currently doing (discrete)
obs(2) = Simulink.BusElement;
obs(2).Name = "isDoingB";

% Vector with the time for each task for the robot between 0 and max_time
obs(3) = Simulink.BusElement;
obs(3).Name = "timeA";
obs(3).Dimensions = [1 num_tasks];
obs(3).Min = 0;
obs(3).Max = max_time;

% Vector with the time for each task for the human between 0 and max_time
obs(4) = Simulink.BusElement;
obs(4).Name = "timeB";
obs(4).Dimensions = [1 num_tasks];
obs(4).Min = 0;
obs(4).Max = max_time;

% Vector with the ergonomic value for each task between 0 and max_erg
obs(5) = Simulink.BusElement;
obs(5).Name = "erg";
obs(5).Dimensions = [1 num_tasks];
obs(5).Min = 0;
obs(5).Max = max_erg;

% Matrix with the dependencies of the tasks to meet assembly constraints
obs(6) = Simulink.BusElement;
obs(6).Name = "dependencies";
obs(6).Dimensions = [num_tasks num_tasks];
obs(6).Min = 0;
obs(6).Max = 1;

% Create bus object
obsBus = Simulink.Bus();
obsBus.Elements = obs;
obsInfo = bus2RLSpec("obsBus","DiscreteElements",{"isDoingA",possible_actions,"isDoingB",possible_actions});

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Create the environment
mdl = "Simulink_environment";
open_system(mdl);
blks = [mdl+"/Robot", mdl+"/Human"];
obsInfos = {obsInfo,obsInfo};
actInfos = {actInfo,actInfo};
env = rlSimulinkEnv(mdl,blks,obsInfos,actInfos);
env.ResetFcn = @(in) resetEnvironment(in);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Create the agents representing robot and human
agentOptions = rlPPOAgentOptions(...
    ExperienceHorizon = 512,...
    ClipFactor = 0.2,...
    EntropyLossWeight = 0.7,...
    MiniBatchSize = 128,...
    NumEpoch = 4,...
    AdvantageEstimateMethod = "gae",...
    GAEFactor = 0.95,...
    SampleTime = 0.1,...
    DiscountFactor = 1);
agentOptions.ActorOptimizerOptions.LearnRate  = 1e-4;
agentOptions.CriticOptimizerOptions.LearnRate = 1e-4;

robot = rlPPOAgent(obsInfo,actInfo, ...
    rlAgentInitializationOptions(NumHiddenUnit= 256),agentOptions);
human = rlPPOAgent(obsInfo,actInfo, ...
    rlAgentInitializationOptions(NumHiddenUnit= 256),agentOptions);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Validate the environment (optional)
% validateEnvironment(env)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Training of the agents
if doTraining
    % Log the training data in the end of each episode in the folder "logs"
    lgr = rlDataLogger();
    lgr.EpisodeFinishedFcn = @logExperienceData;
    lgr.LoggingOptions.LoggingDirectory = "logs";

    % Define training options
    trainOpts = rlMultiAgentTrainingOptions(...
        AgentGroups = "auto",...
        LearningStrategy = "decentralized",...
        MaxEpisodes = 800,...
        MaxStepsPerEpisode = 1400,...
        ScoreAveragingWindowLength = 30,...
        StopTrainingCriteria = "EpisodeReward",...
        StopTrainingValue = [1000 1000]);

    % Start the training
    training_results = train([robot,human],env,trainOpts,logger=lgr);

    % Change of parameters for the second stage of training (optional)
    % robot.AgentOptions.EntropyLossWeight = 0.01;
    % human.AgentOptions.EntropyLossWeight = 0.01;
    % training_results(1, 1).TrainingOptions.MaxEpisodes = 1300;
    % training_results(1, 2).TrainingOptions.MaxEpisodes = 1300;
    % training_results = train([robot,human],env,training_results,logger=lgr);

    % Save trained agents and training results
    save("Robot","robot")
    save("Human","human")
    save("Training-results","training_results")
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Load trained agents and simulate assembly process
if ~doTraining
    % Robot-agent and human-agent of the same case should be selected
    file1 = uigetfile("Trained_agents\");
    file2 = uigetfile("Trained_agents\");
    load("Trained_agents\"+file1)
    load("Trained_agents\"+file2)
    out = {};
    reward=[];
    for i = 1:10
        out{i} = sim([robot,human],env); % Simulation of the assembly process
        time(i) = out{1, i}(1).IsDone.Time(end) + 0.1;
    end
    time_min = find(time == min(time));
    ind = time_min(1);

    % The task time -1 is shown, every task actually starts 1 second before
    figure
    tasksRobot = out{1,ind}(1).Observation.isDoingA.plot;
    ylim([0 num_tasks+1]);
    title("Robot"); xlabel("Time"); ylabel("Task");
    figure
    tasksHuman = out{1,ind}(1).Observation.isDoingB.plot;
    ylim([0 num_tasks+1]);
    title("Human"); xlabel("Time"); ylabel("Task");
end